{- Mateusz Styrczula -}


-- For every formula in a propositional calculus (sentential calculus)
-- equivalent formula in Conjunctive Normal Form exists.


module CNF_converter where


import Formula_lexer
import Formula_parser


-- eliminate (<=>) and (=>) from given formula
eliminate :: (Eq a, Show a) => PT a -> PT a
eliminate (t1 `EQU` t2) = ((elt1 `AND` elt2) `OR` ((NOT elt1) `AND` (NOT elt2)))
  where
    elt1 = eliminate t1
    elt2 = eliminate t2
eliminate (t1 `IMP` t2) = ((NOT (eliminate t1)) `OR` (eliminate t2))
eliminate (t1 `OR` t2)  = ((eliminate t1) `OR` (eliminate t2))
eliminate (t1 `AND` t2) = ((eliminate t1) `AND` (eliminate t2))
eliminate (NOT t)       = (NOT (eliminate t))
eliminate t             = t


-- using de Morgan's and double-negation laws move all negations "inside"
demorgan :: (Eq a, Show a) => PT a -> PT a
demorgan (NOT (t1 `AND` t2)) = ((demorgan (NOT t1)) `OR` (demorgan (NOT t2)))
demorgan (NOT (t1 `OR` t2))  = ((demorgan (NOT t1)) `AND` (demorgan (NOT t2)))
demorgan (t1 `AND` t2)       = ((demorgan t1) `AND` (demorgan t2))
demorgan (t1 `OR` t2)        = ((demorgan t1) `OR` (demorgan t2))
demorgan (NOT (NOT t))       = demorgan t
demorgan t                   = t


-- A v (B ^ C) = (A v B) ^ (A v C)
-- (A ^ B) v C = (A v C) ^ (B v C)
distribute :: (Eq a, Show a) => PT a -> PT a
distribute (a `OR` (b `AND` c)) = distribute ((distribute (a `OR` b)) `AND` (distribute (a `OR` c)))
distribute ((a `AND` b) `OR` c) = distribute ((distribute (a `OR` c)) `AND` (distribute (b `OR` c)))
distribute (a `OR` b)           =
    let
      dista = distribute a
      distb = distribute b
      is_dista_in_and_form = case dista of
                               (_ `AND` _) -> True
                               _           -> False
      is_distb_in_and_form = case distb of
                               (_ `AND` _) -> True
                               _           -> False
    in
      if is_dista_in_and_form || is_distb_in_and_form
        then distribute (dista `OR` distb)
        else (dista `OR` distb)
distribute (a `AND` b)          = ((distribute a) `AND` (distribute b))
distribute t                    = t


-- "main" converting function
cnf :: (Eq a, Show a) => PT a -> PT a
cnf t = distribute (demorgan (eliminate t))
