{- Mateusz Styrczula -}


-- For every formula in a propositional calculus (sentential calculus)
-- equivalent formula in Clause Form exists.


module ClauseForm_converter where


import Formula_lexer
import Formula_parser
import qualified Data.Set as Set


-- datatype representing Clause Literals
data Literal a =
    V a
  | N a
instance (Show a) => Show (Literal a)
  where
    show (V a) = show a
    show (N a) = "~" ++ (show a)
instance (Eq a) => Eq (Literal a)
  where
    V a == V b = a == b
    N a == N b = a == b
    _ == _     = False
instance (Ord a) => Ord (Literal a)
  where
    (V a) `compare` (V b) = a `compare` b
    (V a) `compare` (N b) = if a `compare` b == EQ then LT else a `compare` b
    (N a) `compare` (V b) = if a `compare` b == EQ then GT else a `compare` b
    (N a) `compare` (N b) = a `compare` b


-- returns true if given literals clashes
infix 4 =!=
(=!=) :: (Eq a) => Literal a -> Literal a -> Bool
(V a) =!= (N b) = a == b
(N a) =!= (V b) = a == b
_ =!= _         = False


-- negate given literal
neg :: (Ord a, Eq a, Show a) => Literal a -> Literal a
neg (V l) = (N l)
neg (N l) = (V l)


-- returns true if given literal clashes with some literal in clause
infix 4 =!=<
(=!=<) :: (Ord a, Eq a, Show a) => Literal a -> Set.Set (Literal a) -> Bool
l =!=< c = Set.member (neg l) c


-- convert one alternative to clause (clause = set of literals)
clause :: (Ord a, Eq a, Show a) => PT a
                                -> Set.Set (Literal a)
                                -> Set.Set (Literal a)
clause (VAR v) c = Set.insert (V v) c
clause (NOT (VAR v)) c = Set.insert (N v) c
clause (t1 `OR` t2) c = (clause t1 c) `Set.union` (clause t2 c)


-- convert Sentence in CNF (stored in "PT Tree") to Set of Clauses (clause = set of literals)
to_cf :: (Ord a, Eq a, Show a) => PT a
                               -> Set.Set (Set.Set (Literal a))
                               -> Set.Set (Set.Set (Literal a))
to_cf (t1 `AND` t2) cs = (to_cf t1 cs) `Set.union` (to_cf t2 cs)
to_cf (c @ (t1 `OR` t2)) cs = Set.insert (clause c Set.empty) cs
to_cf (c @ (VAR v)) cs = Set.insert (clause c Set.empty) cs
to_cf (c @ (NOT (VAR v))) cs = Set.insert (clause c Set.empty) cs


-- "main" converting function
cf :: (Ord a, Eq a, Show a) => PT a -> Set.Set (Set.Set (Literal a))
cf sentence = to_cf sentence Set.empty
