{- Mateusz Styrczula -}


-- Formula LEXER:


module Formula_lexer where


-- Tokens produced by lexer (used by parser)
data Token a = T_EQU | T_IMP | T_OR | T_AND | T_NOT | T_LB | T_RB | T_VAR a | T_END | T_ERROR String
instance (Show a) => Show (Token a)
  where
    show T_EQU       = "<=>"
    show T_IMP       = "=>"
    show T_OR        = "v"
    show T_AND       = "^"
    show T_NOT       = "~"
    show T_LB        = "("
    show T_RB        = ")"
    show (T_VAR s)   = show s
    show T_END       = "END"
    show (T_ERROR s) = "err: " ++ s
instance (Eq a) => Eq (Token a)
  where
    T_EQU == T_EQU             = True
    T_IMP == T_IMP             = True
    T_OR == T_OR               = True
    T_AND == T_AND             = True
    T_NOT == T_NOT             = True
    T_LB == T_LB               = True
    T_RB == T_RB               = True
    (T_VAR a) == (T_VAR b)     = a == b
    T_END == T_END             = True
    (T_ERROR _) == (T_ERROR _) = True
    _ == _                     = False


-- small "v" isn't a letter!
letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuwxyz"
digits = "0123456789"
white = " \t\n"


-- get one token from string
get_token :: String -> (Token String, String)
get_token [] = (T_END, [])
get_token (chars @ (c:cs))
    | c == '~' = (T_NOT, cs)
    | c == '^' = (T_AND, cs)
    | c == 'v' = (T_OR, cs)
    | c == '(' = (T_LB, cs)
    | c == ')' = (T_RB, cs)
    | c == '=' = case cs of
                   []        -> (T_ERROR "unexpected end", [])
                   ('>':ncs) -> (T_IMP, ncs)
                   _         -> (T_ERROR "bad token", [])
    | c == '<' = case cs of
                   []            -> (T_ERROR "unexpected end", [])
                   ('=':'>':ncs) -> (T_EQU, ncs)
                   _             -> (T_ERROR "bad token", [])
    | c `elem` white                         = get_token cs
    | c `elem` digits || c `notElem` letters = (T_ERROR "bad token", [])
    | otherwise                              = (T_VAR identifier, rest)
  where
    read_word [] acc = (reverse acc, [])
    read_word (chars @ (c:cs)) acc
        | c `elem` (letters ++ digits) = read_word cs (c:acc)
        | otherwise                    = (reverse acc, chars)
    (identifier, rest) = read_word chars []
 

-- convert string to token list
tokenize :: String -> [Token String]
tokenize [] = [T_END]
tokenize chars =
    let
      (token, rest) = get_token chars
    in
      token : (tokenize rest)
