{- Mateusz Styrczula -}


-- Formula PARSER:
{-
  Grammar in EBNF:
         parse_equ = parse_imp [{"<=>" parse_imp}]
         parse_imp = parse_or ["=>" parse_imp]
          parse_or = parse_and [{"v" parse_and}]
         parse_and = parse_not [{"^" parse_not}]
         parse_not = [{"~"}] parse_var
         parse_var = T_VAR | "(" parse_equ ")"
-}


module Formula_parser where


import Formula_lexer


-- Tree containing parsed sentence - produced by parser
data PT a =
      VAR a
    | NOT (PT a)
    | (PT a) `AND` (PT a)
    | (PT a) `OR` (PT a)
    | (PT a) `IMP` (PT a)
    | (PT a) `EQU` (PT a)
instance (Show a) => Show (PT a)
  where
    show (VAR a) = show a
    show (NOT pt) = "~" ++ (show pt)
    show (l `AND` r) = "(" ++ show l ++ " ^ " ++ show r ++ ")"
    show (l `OR` r) = "(" ++ show l ++ " v " ++ show r ++ ")"
    show (l `IMP` r) = "(" ++ show l ++ " => " ++ show r ++ ")"
    show (l `EQU` r) = "(" ++ show l ++ " <=> " ++ show r ++ ")"


-- I have created this function to avoid left-side recursion.
-- ex: x ^ x ^ x ^ x  ===  [x_tree, x_tree, x_tree, x_tree]
-- Output from this function is used to build bigger tree with "left" fixity.
multi_parse :: (Eq a, Show a) => Token a
                              -> ([Token a] -> (PT a, [Token a]))
                              -> [Token a]
                              -> ([PT a], [Token a])
multi_parse tok parser ts =
  let
    (revtreelist, rest) = mp ts []
  in
    (reverse revtreelist, rest)
  where
    mp tss acc =
      let 
        (subt, restts) = parser tss 
      in
        case restts of
          (t:rest) -> if t == tok
                        then mp rest (subt:acc)
                        else (subt:acc, restts)
          _        -> (subt:acc, restts)


-- parse_var = T_VAR | "(" parse_equ ")"
parse_var :: (Eq a, Show a) => [Token a] -> (PT a, [Token a])
parse_var ((T_VAR x):ts) = (VAR x, ts)
parse_var (T_LB:ts) =
    case restts of
      (T_RB:rest)       -> (subt, rest)
      ((T_ERROR msg):_) -> error ("parse error: " ++ msg)
      _                 -> error "parse error: cannot find closing bracket..."
  where
    (subt, restts) = parse_equ ts
parse_var ((T_ERROR msg):_) = error ("parse error: " ++ msg)
parse_var (t:_) = error ("parse error: got " ++ show t ++ ", VAR expected...")
parse_var [] = error "parse_var: got [] - this is REALLY impossible!"


-- parse_not = [{"~"}] parse_var
parse_not :: (Eq a, Show a) => [Token a] -> (PT a, [Token a])
parse_not (T_NOT:ts) = (NOT subt, restts)
  where
    (subt, restts) = parse_not ts
parse_not ts = parse_var ts


-- parse_and = parse_not [{"^" parse_not}]
parse_and :: (Eq a, Show a) => [Token a] -> (PT a, [Token a])
parse_and ts =
  let
    (subtree_list, rest) = multi_parse T_AND parse_not ts
    proper_tree = foldl1 (\acc x -> acc `AND` x) subtree_list
  in
    (proper_tree, rest)


-- parse_or = parse_and [{"v" parse_and}]
parse_or :: (Eq a, Show a) => [Token a] -> (PT a, [Token a])
parse_or ts =
  let
    (subtree_list, rest) = multi_parse T_OR parse_and ts
    proper_tree = foldl1 (\acc x -> acc `OR` x) subtree_list
  in
    (proper_tree, rest)


-- parse_imp = parse_or ["=>" parse_imp]
parse_imp :: (Eq a, Show a) => [Token a] -> (PT a, [Token a])
parse_imp ts = 
    case restl of
      (T_IMP:rest) -> let 
                        (subrt, restr) = parse_imp rest
                      in
                        (sublt `IMP` subrt, restr)
      _            -> (sublt, restl)
  where
    (sublt, restl) = parse_or ts


-- parse_equ = parse_imp [{"<=>" parse_imp}]
parse_equ :: (Eq a, Show a) => [Token a] -> (PT a, [Token a])
parse_equ ts =
  let
    (subtree_list, rest) = multi_parse T_EQU parse_imp ts
    proper_tree = foldl1 (\acc x -> acc `EQU` x) subtree_list
  in
    (proper_tree, rest)


-- "main" parser function
parse :: (Eq a, Show a) => [Token a] -> PT a
parse ts =
    case rest of
      [T_END]         -> tree
      [(T_ERROR msg)] -> error ("parse error: " ++ msg)
      _               -> error "parse error: syntactically bad formula"
  where
    (tree, rest) = parse_equ ts
