{- Mateusz Styrczula -}


-- little helper (disturber??) type definition


module NumberedClauses where


import Formula_lexer
import Formula_parser
import ClauseForm_converter
import qualified Data.Set as Set
import qualified Data.List as List
import qualified Data.Map as Map


-- clause = (self_nr, set of literals, mum_nr, dad_nr)
newtype NumberedClause a = NC (Int, Set.Set (Literal a), Int, Int)
instance (Eq a) => Eq (NumberedClause a)
  where
    (NC (_,x,_,_)) == (NC (_,y,_,_)) = x == y
instance (Ord a) => Ord (NumberedClause a)
  where
    (NC (_,x,_,_)) `compare` (NC (_,y,_,_)) = x `compare` y
instance (Show a) => Show (NumberedClause a)
  where
    show (NC nc) = show nc


-- extract NumberedClause mum number
husk_mum :: (Ord a, Eq a, Show a) => NumberedClause a -> Int
husk_mum (NC (_,_,mum,_)) = mum


-- extract NumberedClause dad number
husk_dad :: (Ord a, Eq a, Show a) => NumberedClause a -> Int
husk_dad (NC (_,_,_,dad)) = dad


-- extract NumberedClause number
husk_nr :: (Ord a, Eq a, Show a) => NumberedClause a -> Int
husk_nr (NC (nr,_,_,_)) = nr


-- return number of literals in clause
clause_size :: (Ord a, Eq a, Show a) => NumberedClause a -> Int
clause_size (NC (_,c,_,_)) = Set.size c


-- return list of literals
literals_list :: (Ord a, Eq a, Show a) => NumberedClause a -> [Literal a]
literals_list (NC (_,c,_,_)) = Set.elems c


-- return set of literals
literals_set :: (Ord a, Eq a, Show a) => NumberedClause a -> Set.Set (Literal a)
literals_set (NC (_,c,_,_)) = c


-- transform clauses to >>NC (self_nr, clause, mum_nr, dad_nr)<< tuple form
-- all initial clauses have mum_nr = dad_nr = 0
transform :: (Ord a, Eq a, Show a) => Set.Set (Set.Set (Literal a))
                                   -> Set.Set (NumberedClause a)
transform cs
    | cs == Set.empty = Set.empty
    | otherwise       =
        Set.fromList (foldl (\(acc @ ((NC (i, _, _, _)):_)) x -> (NC (i+1, x, 0, 0)):acc) 
                            [NC (1, head clause_list, 0, 0)]
                            (tail clause_list))
  where
    clause_list = Set.elems cs


-- transform Set to Map
peel :: (Ord a, Eq a, Show a) => Set.Set (NumberedClause a)
                              -> Map.Map Int (Set.Set (Literal a), Int, Int)
peel cs = p clause_list Map.empty
  where
    clause_list = Set.elems cs
    p [] s = s
    p ((NC (n,c,m,d)):clauses) s = p clauses (Map.insert n (c,m,d) s)
