{- Mateusz Styrczula -}


-- My own (I suppose rather weak, inefficient and moreover not very brilliant)
-- implementation of "Sentential Calculus Theorem Proover"
-- using "Resolution Method" formulated by J.A.Robinson in 1965


module Robinson where


import Formula_lexer
import Formula_parser
import CNF_converter
import ClauseForm_converter
import NumberedClauses
import qualified Data.Set as Set
import qualified Data.List as List
import qualified Data.Map as Map
import Debug.Trace
import Text.Printf


-- returns True if given clause (set of literals) is True
-- (clause is considered True when there are at least two literals in clash)
is_clause_true :: (Ord a, Eq a, Show a) => Set.Set (Literal a) -> Bool
is_clause_true c = c_elems /= nubbed
  where
    c_elems = Set.elems (c)
    nubbed = List.nubBy (=!=) c_elems


-- removes true clauses from set
remove_true_clauses :: (Ord a, Eq a, Show a) => Set.Set (Set.Set (Literal a))
                                             -> Set.Set (Set.Set (Literal a))
remove_true_clauses cs = Set.filter (not . is_clause_true) cs




-- yes... this is big, ugly, monolithic function-monster...
-- I guess that this is write-only form on human-computer interaction (like APL ;)...
resolution :: (Ord a, Eq a, Show a) => Set.Set (NumberedClause a)
                                    -> (Bool, Set.Set (NumberedClause a))
resolution cs
    | cs == Set.empty                      =
        trace "Clause set doesn't contain empty clause (it is empty in fact), so IT IS NOT CONTRADICTORY!"
              (True, cs)
    | Set.member (NC (0,Set.empty,0,0)) cs =
        trace "Clause set contains empty clause, so IT IS CONTRADICTORY!"
              (True, cs)
    | otherwise                            = resol clause_list

  where

--    clause_list = List.sortBy (\a b -> if clause_size (a) > clause_size (b) then GT else LT) (Set.elems cs)
    clause_list = Set.elems cs

    -- ...
    resol [] =
      trace "Clause set doesn't contain empty clause, so IT IS NOT CONTRADICTORY!"
            (True, cs)
    resol (clause:clauses) =
      let
        (ended, final_clause_set) = check_clash clause (literals_list clause) clause_list
      in
        if ended
          then (ended, final_clause_set)
          else resol clauses
    
    -- ...
    check_clash cl [] _ = (False, cs)
    check_clash cl (c_lit @ (literal:literals)) cl_lst =
      case find_clashing_clause literal cl_lst of
        Nothing                       -> check_clash cl literals clause_list
        Just (ccl, unchecked_clauses) ->
          let
            c1 = Set.delete literal (literals_set cl)
            c2 = Set.delete (neg literal) (literals_set ccl)
            resolvent = c1 `Set.union` c2
            resolvent_nc = NC ((Set.size cs) + 1, resolvent, husk_nr cl, husk_nr ccl)
          in
            if (is_clause_true resolvent) || (resolvent_nc `Set.member` cs)
              then check_clash cl c_lit unchecked_clauses
              else trace (show resolvent_nc)
                         (resolution (Set.insert resolvent_nc cs))

    -- ...
    find_clashing_clause l [] = Nothing
    find_clashing_clause l ((cl @ (NC (_,set_of_literals,_,_))):cls)
        | l =!=< set_of_literals = Just (cl, cls)
        | otherwise              = find_clashing_clause l cls





-- perform [string] -> [token] conversion, build abstract tree (parse),
-- convert to Conjuctive Normal Form, build set of clauses,
-- then perform prove by resolution and finally
-- (via "backtracking") build proof list
prove :: String -> ([(Int, (Set.Set (Literal String), Int, Int))],
                    [(Int, (Set.Set (Literal String), Int, Int))])
prove sentence
    | not contradictory = (Map.assocs pure_cf_map, [])
    | otherwise         = (Map.assocs pure_cf_map, p)
  where
    pure_cf = transform (remove_true_clauses (cf (cnf (NOT (parse (tokenize (sentence)))))))
    pure_cf_map = peel pure_cf
    (_, cf_after) = trace "Trace:" (resolution pure_cf)
    contradictory = Set.member (NC (0,Set.empty,0,0)) cf_after
    cf_after_map = peel cf_after
    cf_after_size = Set.size cf_after
    build_proof map (proof @ ((_,(_,mum_nr,dad_nr)):rest_of_proof))
        | mum_nr == 0 || dad_nr == 0 = rest_of_proof
        | otherwise                  =
            let
              mum = (mum_nr, map Map.! mum_nr)
              dad = (dad_nr, map Map.! dad_nr)
              by_mum = build_proof map (mum:proof)
              by_dad = build_proof map (dad:proof)
            in
              by_mum `List.union` by_dad
    p = List.sort (build_proof cf_after_map [(cf_after_size, cf_after_map Map.! cf_after_size)])





-- -- misc "pretty-printing" garbage below ---------------------------------------


-- ...
find_longest_clause :: [(Int, (Set.Set (Literal String), Int, Int))] -> Int -> Int
find_longest_clause [] acc = acc
find_longest_clause ((_, (clause, _, _)):rest) acc
    | clause_len > acc = find_longest_clause rest clause_len
    | otherwise        = find_longest_clause rest acc
  where
    clause_len = length (show clause)


-- convert proof list to string
pretty :: [(Int, (Set.Set (Literal String), Int, Int))] -> Int -> String
pretty [] _ = ""
pretty ((nr, (clause, mum, dad)):rest) just_col = 
    "  " ++ (show nr) ++ "\t" ++ clause_s 
         ++ (if mum == 0 && dad == 0
               then "\n"
               else gen_spaces (just_col-clause_len) ++ " <- " ++
                    (show mum) ++ "," ++ (show dad) ++ "\n")
         ++ (pretty rest just_col)
  where
    clause_s = show clause
    clause_len = length clause_s
    gen_spaces n
        | n <= 0    = ""
        | otherwise = ' ' : gen_spaces (n-1)


-- Finally ...
pretty_prove :: String -> IO ()
pretty_prove s = do
    (cf, proof) <- return $! (prove s)
    cf_s <- return $! (pretty cf 0)
    proof_s <- return $! (pretty proof (find_longest_clause proof 0))

    printf "\n--------------------------------------\n"
    printf "Formula:\n\n"
    printf "%s\n\n" s
    printf "Clause set equivalent to formula's negation:\n\n"
    printf "%s\n" cf_s
    printf "Proof:\n\n"
    printf "%s" proof_s
    printf "--------------------------------------\n"
